@@ -6,7 +6,7 @@ import GPy
 import itertools
 
 from cwgp.kernel import RBF, OU, Matern32
-from cwgp.transformations import sal, sa, asinh, box_cox, inv_sal, inv_sa, inv_asinh, inv_box_cox
+from cwgp.transformations import sal, sa, asinh, box_cox, affine, inv_sal, inv_sa, inv_asinh, inv_box_cox, inv_affine, d_sal, d_sa, d_asinh, d_box_cox, d_affine
 
 
 class Phi():
@@ -14,37 +14,51 @@ class Phi():
         "sal": {
             "fn": sal,
             "inv_fn": inv_sal,
+            "d_fn": d_sal,
             "par_len": 4,
             "bounds": [
-                (-20,
-                 20)] * 4},
+                (None,
+                 None)] * 4},
         "sa": {
             "fn": sa,
             "inv_fn": inv_sa,
+            "d_fn": d_sa,
             "par_len": 2,
             "bounds": [
-                (-20,
-                 20)] * 2},
+                (None,
+                 None)] * 2},
         "asinh": {
             "fn": asinh,
             "inv_fn": inv_asinh,
+            "d_fn": d_asinh,
             "par_len": 4,
             "bounds": [
-                (-20,
-                 20),
+                (None,
+                 None),
                 (1e-5,
                  None),
                 (1e-5,
                  None),
-                (-20,
-                 20)]},
+                (None,
+                 None)]},
         "box_cox": {
             "fn": box_cox,
             "inv_fn": inv_box_cox,
+            "d_fn": d_box_cox,
             "par_len": 1,
             "bounds": [
                 (1e-5,
                  None)]},
+        "affine": {
+            "fn": affine,
+            "inv_fn": inv_affine,
+            "d_fn": d_affine,
+            "par_len": 2,
+            "bounds": [
+                (None,
+                 None),
+                (None,
+                 None)]},
     }
 
     KERNEL_BANK = {
@@ -56,31 +70,31 @@ class Phi():
     def __init__(
             self,
             fn,
-            kernel="RBF",
-            kernel_params_estimate=True,
-            ARD=False,
             **kwargs
     ):
         self.fn = [self.FN_BANK[f]["fn"]
                    for f in fn]  # a differentiable function
         self.inv_fn = [self.FN_BANK[f]["inv_fn"] for f in fn[::-1]]
         self.par_len = [self.FN_BANK[f]["par_len"] for f in fn]
-        self.d_fn = [elementwise_grad(f, 1)
-                     for f in self.fn]  # take derivative
-        self.kernel = self.KERNEL_BANK[kernel]["kern"]
-        self.kernel_name = kernel
-        self.kernel_params = self.KERNEL_BANK[kernel]["params"] if kernel_params_estimate else 0
-        self.init_scale = self.KERNEL_BANK[kernel]["init_scale"]
-        self.ARD = ARD
+        # self.d_fn = [elementwise_grad(f, 1)
+        #              for f in self.fn]  # take derivative
+        self.d_fn = [self.FN_BANK[f]["d_fn"]
+                     for f in fn]
+        self.kernel_name = kwargs.get("kernel", "RBF")
+        self.kernel = self.KERNEL_BANK[self.kernel_name]["kern"]
+        self.kernel_params = self.KERNEL_BANK[self.kernel_name]["params"] if kwargs.get(
+            "kernel_params_estimate", True) else 0
+        self.init_scale = self.KERNEL_BANK[self.kernel_name]["init_scale"]
+        self.ARD = kwargs.get("ARD", False)
         self.bounds = list(itertools.chain.from_iterable(
-            [self.FN_BANK[f]["bounds"] for f in fn])) + [(None, None)] * self.kernel_params
+            [self.FN_BANK[f]["bounds"] for f in fn])) + [(1e-5, None)] * self.kernel_params
 
     def comp_phi(self, par, y):
         assert len(par) >= sum(self.par_len), "Not enough parameters"
         comp = copy.deepcopy(y)
-        d_comp, par_len = 1, 0
+        d_comp, par_len = 0, 0
         for i in range(0, len(self.fn)):
-            d_comp *= self.d_fn[i](par[par_len:], comp)
+            d_comp += np.log(self.d_fn[i](par[par_len:], comp))
             comp = self.fn[i](par[par_len:], comp)
             par_len += self.par_len[i]
         return comp, d_comp
@@ -97,7 +111,7 @@ class Phi():
         return inv_comp
 
     def likelihood(self, par, y, t, mf):
-        phi_y, chain_d_sal = self.comp_phi(par, y)
+        phi_y, d_phi_y = self.comp_phi(par, y)
         t_phi_y = np.transpose(phi_y)
         t_t = np.transpose(t)
         mean_t = mf(t) if mf else np.zeros(t.shape)
@@ -108,9 +122,10 @@ class Phi():
             cov_xx = self.kernel(1).K(t.reshape(-1, 1), t.reshape(-1, 1))
 
         gaussian_params = 0.5 * \
-            (t_t - np.transpose(mean_t)) @ np.linalg.inv(cov_xx) @ (t - mean_t)
-        return np.ravel(0.5 * np.log(np.linalg.det(cov_xx)) +
-                        gaussian_params - np.sum(np.log(chain_d_sal)))
+            (t_phi_y - np.transpose(mean_t)) @ np.linalg.inv(cov_xx) @ (phi_y - mean_t)
+
+        sign, logdet = np.linalg.slogdet(cov_xx)
+        return 0.5 * sign * logdet + gaussian_params - np.sum(d_phi_y)
 
     def reml(self, method):
         pass
@@ -129,7 +144,6 @@ class Phi():
                 try:
                     res = minimize(
                         self.likelihood,
-
                         np.random.uniform(0, self.init_scale,
                                           sum(self.par_len) + self.kernel_params),
                         args=(
